/*------------------------------------------------------------------------*/
/* IML-BKP.SAS - SAS/IML implementation of a standard three-layer         */
/*               back-propagation network                                 */
/*------------------------------------------------------------------------*/
/* Copyright (C) 1995-2001 Chih-Hao Tsai                                  */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or (at  */
/* your option) any later version.                                        */
/*                                                                        */
/* This program is distributed in the hope that it will be useful, but    */
/* WITHOUT ANY WARRANTY; without even the implied warranty of             */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      */
/* General Public License for more details.                               */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, */
/* USA.                                                                   */
/*------------------------------------------------------------------------*/
/* Program: IML-BKP.SAS                                                   */
/* Purpose: SAS/IML implementation of a standard three-layer              */
/*          back-propagation network                                      */
/* Date of first version: March 18, 1995                                  */
/* This version: 0.9 (January 23, 1999)                                   */
/* First public release: May 2, 2001                                      */
/* Author: Chih-Hao Tsai                                                  */
/* Email: hao520@yahoo.com                                                */
/* IML-BKP website: http://www.geocities.com/hao520/research/             */
/* Author's homepage: http://www.geocities.com/hao520                     */
/*........................................................................*/
/*  n_input: (parameter) number of input units                            */
/* n_output: (parameter) number of output units                           */
/* n_hidden: (parameter) number of hidden units                           */
/*  n_epoch: (parameter) number of epoches                                */
/* n_pattrn: (parameter) number of patterns per epoch                     */
/*  wtrange: (parameter) range of initial random weights                  */
/*  epsilon: (scalar) learning rate                                       */
/* bepsilon: (scalar) bias unit learning rate                             */
/* momentum: (scalar) momentum parameter                                  */
/*     bias: (scalar) bias input constant                                 */
/*           .............................................................*/
/*   PATTRN: (matrix) all input patterns                                  */
/*  TEACHER: (matrix) all target patterns                                 */
/* inpattrn: (vector) single input pattern                                */
/*           .............................................................*/
/* veti_hdn: (vector) hidden, net inputs                                  */
/*  act_hdn: (vector) hidden, activations                                 */
/*    error: (vector) output, (target - activation)                       */
/*      pss: (vector) output, pattern sum of squared errors               */
/*      tss: (vector) output, total sum of squared errors                 */
/*  dlta_hi: (vector) hidden, generalized deltas                          */
/*    WT_HI: (matrix) hidden<->input, weights                             */
/*   DWT_HI: (matrix) hidden<->input, delta weights                       */
/*  DWTM_HI: (matrix) hidden<->input, delta weights at time t-1           */
/*   WED_HI: (matrix) hidden<->input, weight error derivatives            */
/*  bias_hb: (vector) hidden<->bias, weights                              */
/*  dbsm_hb: (vector) hidden<->bias, delta weights at time t-1            */
/*   bed_hb: (vector) hidden<->bias, weight error derivatives             */
/* dbias_hb: (vector) hidden<->bias, delta weights                        */
/*           .............................................................*/
/* neti_out: (vector) output, net inputs                                  */
/*  act_out: (vector) output, activations                                 */
/*  dlta_oh: (vector) output, generalized deltas                          */
/*    WT_OH: (matrix) output<->hidden, weights                            */
/*   DWT_OH: (matrix) output<->hidden, delta weights                      */
/*  DWTM_OH: (matrix) output<->hidden, delta weights at time t-1          */
/*   WED_OH: (matrix) output<->hidden, weight error derivatives           */
/*  bias_ob: (vector) output<->bias, weights                              */
/*  dbsm_ob: (vector) output<->bias, delta weights at time t-1            */
/*   bed_ob: (vector) output<->bias, weight error derivatives             */
/* dbias_ob: (vector) output<->bias, delta weights                        */
/*........................................................................*/
/*    crate: (scalar) cascade rate (1=turn off cascade mode)              */
/* netm_hdn: (vector) hidden, net inputs at time t-1                      */
/* netm_out: (vector) output, net inputs at time t-1                      */
/*------------------------------------------------------------------------*/

options linesize=78 pagesize=58 nodate pageno=1;

proc iml;

reset noprint;
reset autoname;

start bkp;

    /* Basic parameters */
n_epoch=0;
n_pattrn=8;
n_input=8;
n_output=8;
n_hidden=3;
epsilon=.1;
bepsilon=.1;
bias=.2;
momentum=.8;
crate=1;
wtrange=.1; /* -.1 -- +.1 */

    /* Initialize network */
sequence=shape(0, n_pattrn, 1);
wt_hi=shape(0,n_hidden,n_input);
wt_oh=shape(0,n_output,n_hidden);
dwtm_hi=shape(0,n_hidden,n_input);
dwtm_oh=shape(0,n_output,n_hidden);
bias_hb=shape(0,n_hidden,1);
bias_ob=shape(0,n_output,1);
dbsm_hb=shape(0,n_hidden,1);
dbsm_ob=shape(0,n_output,1);
netm_hdn=shape(0,n_hidden,1);
netm_out=shape(0,n_output,1);

do i=1 to n_pattrn;
  sequence[i]=i;
end;

    /* Initialize connection weights */
do i=1 to n_hidden;
  do j=1 to n_input;
    wt_hi[i,j]=wtrange-uniform(0)/(1/(wtrange*2));
  end;
end;
do i=1 to n_output;
  do j=1 to n_hidden;
    wt_oh[i,j]=wtrange-uniform(0)/(1/(wtrange*2));
  end;
end;
do i=1 to n_hidden;
  bias_hb[i,1]=wtrange-uniform(0)/(1/(wtrange*2));
end;
do i=1 to n_output;
  bias_ob[i,1]=wtrange-uniform(0)/(1/(wtrange*2));
end;

    /* Begin training */
do until(success=n_pattrn);
  pss=0;
  tss=0;
  success=0;

    /* Shuffle */
  do i=n_pattrn to 2 by -1;
    j=int(uniform(0)*i)+1;
    k=sequence[i];
    sequence[i]=sequence[j];
    sequence[j]=k;
  end;

  do i=1 to n_pattrn;

    /* Get input and target patterns */
    inpattrn=pattern[sequence[i],];
    target=teacher[sequence[i],];

    /* Calculate net input to hidden units */
    neti_hdn=bias_hb#bias+crate#wt_hi*inpattrn`+(1-crate)#netm_hdn;

    /* Store the old net inputs */
    netm_hdn=neti_hdn;

    /* Calculate activation level of hidden units */
    act_hdn=(1+exp(-neti_hdn))##(-1);

    /* Calculate net input to output units */
    neti_out=bias_ob#bias+crate#wt_oh*act_hdn+(1-crate)#netm_out;

    /* Store the old net inputs */
    netm_out=neti_out;

    /* Calculate activation level of output units */
    act_out=(1+exp(-neti_out))##(-1);

    /* Calculate errors */
    error=target`-act_out;

    /* Has the net learned? */
    if abs(error)[<>,]<=.1 then success=success+1;

    /* Calculate generalized deltas for output units */
    dlta_oh=error#act_out#(1-act_out);

    /* Calculate pattern and total sum of squared errors */
    pss=sum(error##2);
    tss=tss+pss;

    /* Calculate generalized deltas for hidden units */
    dlta_hi=act_hdn#(1-act_hdn)#(wt_oh`*dlta_oh);

    /* Calculate output-hidden weight error derivatives */
    wed_oh=dlta_oh*act_hdn`;

    /* Calculate output-hidden delta weights */
    dwt_oh=epsilon#wed_oh+momentum#dwtm_oh;

    /* Store the old output-hidden delta weights */
    dwtm_oh=dwt_oh;

    /* Update the output-hidden weights */
    wt_oh=wt_oh+dwt_oh;

    /* Calculate output-bias weight error derivatives */
    bed_ob=dlta_oh;

    /* Calculate output-bias delta weights */
    dbias_ob=bepsilon#bed_ob+momentum#dbsm_ob;

    /* Store the old output-bias delta weights */
    dbsm_ob=dbias_ob;

    /* Update the output-bias weights */
    bias_ob=bias_ob+dbias_ob;

    /* Calculate hidden-input weight error derivatives */
    wed_hi=dlta_hi*inpattrn;

    /* Calculate hidden-input delta weights */
    dwt_hi=epsilon#wed_hi+momentum#dwtm_hi;

    /* Store the old hidden-input delta weights */
    dwtm_hi=dwt_hi;

    /* Update the hidden-input weights */
    wt_hi=wt_hi+dwt_hi;

    /* Calculate hidden-bias weight error derivatives */
    bed_hb=dlta_hi;

    /* Calculate hidden-bias delta weights */
    dbias_hb=bepsilon#bed_hb+momentum#dbsm_hb;

    /* Store the old hidden-bias delta weights */
    dbsm_hb=dbias_hb;

    /* Update the hidden-bias weights */
    bias_hb=bias_hb+dbias_hb;
  end;
  n_epoch=n_epoch+1;
end;

    /* Print the final weight matrix */
print "Solution Reached at Epoch" n_epoch;
print "Hidden-Input Weight Matrix" wt_hi;
print "Output-Hidden Weight Matrix" wt_oh;

    /* Test the model */

    /* Randomize */
  do i=n_pattrn to 2 by -1;
    j=int(uniform(0)*i)+1;
    k=sequence[i];
    sequence[i]=sequence[j];
    sequence[j]=k;
  end;

  do i=1 to n_pattrn;

  /* Get input and target patterns */
  inpattrn=pattern[sequence[i],];
    target=teacher[sequence[i],];

  /* Calculate net input to hidden units */
  neti_hdn=bias_hb#bias+crate#wt_hi*inpattrn`+(1-crate)#netm_hdn;

  /* Store the old net inputs */
  netm_hdn=neti_hdn;

  /* Calculate activation level of hidden units */
  act_hdn=(1+exp(-neti_hdn))##(-1);

  /* Calculate net input to output units */
  neti_out=bias_ob#bias+crate#wt_oh*act_hdn+(1-crate)#netm_out;

  /* Store the old net inputs */
  netm_out=neti_out;

  /* Calculate activation level of output units */
  act_out=(1+exp(-neti_out))##(-1);

  print "Input Units" inpattrn;
  print "Activation Level of Hidden Units" act_hdn;
  print "Activation Level of Output Units" act_out;
end;
finish bkp;

pattern={1 0 0 0 0 0 0 0,
         0 1 0 0 0 0 0 0,
         0 0 1 0 0 0 0 0,
         0 0 0 1 0 0 0 0,
         0 0 0 0 1 0 0 0,
         0 0 0 0 0 1 0 0,
         0 0 0 0 0 0 1 0,
         0 0 0 0 0 0 0 1};

teacher={1 0 0 0 0 0 0 0,
         0 1 0 0 0 0 0 0,
         0 0 1 0 0 0 0 0,
         0 0 0 1 0 0 0 0,
         0 0 0 0 1 0 0 0,
         0 0 0 0 0 1 0 0,
         0 0 0 0 0 0 1 0,
         0 0 0 0 0 0 0 1};
run bkp;
quit;
